import logging
from datetime import datetime
from typing import Optional, Union

from typing_extensions import Self

from acryl_datahub_cloud.sdk.assertion.assertion_base import (
    AssertionMode,
    _AssertionPublic,
    _HasColumnMetricFunctionality,
    _HasSchedule,
)
from acryl_datahub_cloud.sdk.assertion_input.assertion_input import (
    DEFAULT_DETECTION_MECHANISM,
    DEFAULT_SCHEDULE,
    AssertionIncidentBehavior,
    DetectionMechanism,
    _DetectionMechanismTypes,
)
from acryl_datahub_cloud.sdk.assertion_input.column_metric_assertion_input import (
    ColumnMetricAssertionParameters,
    MetricInputType,
    OperatorInputType,
)
from acryl_datahub_cloud.sdk.entities.assertion import Assertion
from acryl_datahub_cloud.sdk.entities.monitor import Monitor
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import AssertionUrn, CorpUserUrn, DatasetUrn, TagUrn

logger = logging.getLogger(__name__)


class ColumnMetricAssertion(
    _HasColumnMetricFunctionality,
    _HasSchedule,
    _AssertionPublic,
):
    """
    A class that represents a column metric assertion.
    This assertion is used to validate the value of a common field / column metric (e.g. aggregation) such as null count + percentage,
    min, max, median, and more. It uses native source types without AI inference.
    """

    def __init__(
        self,
        *,
        urn: AssertionUrn,
        dataset_urn: DatasetUrn,
        column_name: str,
        metric_type: MetricInputType,
        operator: OperatorInputType,
        # Consolidated criteria parameters
        criteria_parameters: Optional[ColumnMetricAssertionParameters] = None,
        # Standard assertion parameters:
        display_name: str,
        mode: AssertionMode,
        schedule: models.CronScheduleClass = DEFAULT_SCHEDULE,
        incident_behavior: list[AssertionIncidentBehavior],
        detection_mechanism: Optional[
            _DetectionMechanismTypes
        ] = DEFAULT_DETECTION_MECHANISM,
        time_bucketing_strategy: Optional[
            models.AssertionTimeBucketingStrategyClass
        ] = None,
        tags: list[TagUrn],
        created_by: Optional[CorpUserUrn] = None,
        created_at: Union[datetime, None] = None,
        updated_by: Optional[CorpUserUrn] = None,
        updated_at: Optional[datetime] = None,
    ):
        """
        Initialize a column metric assertion.

        Args:
            urn: The URN of the assertion.
            dataset_urn: The URN of the dataset to monitor.
            column_name: The name of the column to monitor.
            metric_type: The type of metric to compute.
            operator: The comparison operator for the metric value.
            criteria_parameters: Optional parameters for the criteria.
            display_name: The display name of the assertion.
            mode: The mode of the assertion (active/inactive).
            schedule: The cron schedule for evaluating the assertion.
            incident_behavior: The behavior when incidents occur.
            detection_mechanism: The mechanism used to detect changes.
            time_bucketing_strategy: Optional time bucketing strategy for data partitioning.
            tags: The tags to apply to the assertion.
            created_by: The URN of the user who created the assertion.
            created_at: The timestamp when the assertion was created.
            updated_by: The URN of the user who last updated the assertion.
            updated_at: The timestamp when the assertion was last updated.
        """
        _AssertionPublic.__init__(
            self,
            urn=urn,
            dataset_urn=dataset_urn,
            display_name=display_name,
            mode=mode,
            tags=tags,
            incident_behavior=incident_behavior,
            detection_mechanism=detection_mechanism,
            time_bucketing_strategy=time_bucketing_strategy,
            created_by=created_by,
            created_at=created_at,
            updated_by=updated_by,
            updated_at=updated_at,
        )
        _HasSchedule.__init__(
            self,
            schedule=schedule,
        )
        _HasColumnMetricFunctionality.__init__(
            self,
            column_name=column_name,
            metric_type=metric_type,
            operator=operator,
            criteria_parameters=criteria_parameters,
        )

    @classmethod
    def _from_entities(cls, assertion: Assertion, monitor: Monitor) -> Self:
        """
        Create a ColumnMetricAssertion from an Assertion and Monitor entity.

        Args:
            assertion: The Assertion entity.
            monitor: The Monitor entity.

        Returns:
            A ColumnMetricAssertion instance.
        """
        return cls(
            urn=assertion.urn,
            dataset_urn=assertion.dataset,
            column_name=cls._get_column_name(assertion),
            metric_type=cls._get_metric_type(assertion),
            operator=cls._get_operator(assertion),
            criteria_parameters=cls._get_criteria_parameters(assertion),
            display_name=assertion.description or "",
            mode=cls._get_mode(monitor),
            schedule=cls._get_schedule(monitor),
            incident_behavior=cls._get_incident_behavior(assertion),
            detection_mechanism=cls._get_detection_mechanism(assertion, monitor),
            time_bucketing_strategy=cls._get_time_bucketing_strategy(monitor),
            tags=cls._get_tags(assertion),
            created_by=cls._get_created_by(assertion),
            created_at=cls._get_created_at(assertion),
            updated_by=cls._get_updated_by(assertion),
            updated_at=cls._get_updated_at(assertion),
        )

    @staticmethod
    def _get_detection_mechanism(
        assertion: Assertion,
        monitor: Monitor,
        default: Optional[_DetectionMechanismTypes] = DEFAULT_DETECTION_MECHANISM,
    ) -> Optional[_DetectionMechanismTypes]:
        """Get the detection mechanism for column metric assertions."""
        parameters = _AssertionPublic._get_validated_detection_context(
            monitor,
            assertion,
            models.AssertionEvaluationParametersTypeClass.DATASET_FIELD,
            models.FieldAssertionInfoClass,
            default,
        )
        if parameters is None:
            return default
        if parameters.datasetFieldParameters is None:
            logger.warning(
                f"Monitor does not have datasetFieldParameters, defaulting detection mechanism to {default}"
            )
            return default
        source_type = parameters.datasetFieldParameters.sourceType
        if source_type == models.DatasetFieldAssertionSourceTypeClass.ALL_ROWS_QUERY:
            additional_filter = _AssertionPublic._get_additional_filter(assertion)
            return DetectionMechanism.ALL_ROWS_QUERY(
                additional_filter=additional_filter
            )
        elif (
            source_type
            == models.DatasetFieldAssertionSourceTypeClass.CHANGED_ROWS_QUERY
        ):
            if parameters.datasetFieldParameters.changedRowsField is None:
                logger.warning(
                    f"Monitor has CHANGED_ROWS_QUERY source type but no changedRowsField, defaulting detection mechanism to {default}"
                )
                return default
            column_name = parameters.datasetFieldParameters.changedRowsField.path
            additional_filter = _AssertionPublic._get_additional_filter(assertion)
            return DetectionMechanism.CHANGED_ROWS_QUERY(
                column_name=column_name, additional_filter=additional_filter
            )
        elif (
            source_type
            == models.DatasetFieldAssertionSourceTypeClass.DATAHUB_DATASET_PROFILE
        ):
            return DetectionMechanism.ALL_ROWS_QUERY_DATAHUB_DATASET_PROFILE
        else:
            logger.warning(
                f"Unsupported DatasetFieldAssertionSourceType {source_type}, defaulting detection mechanism to {default}"
            )
            return default



from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

if TYPE_CHECKING:
    pass

from acryl_datahub_cloud.sdk.assertion_input.assertion_input import (
    DEFAULT_DAILY_SCHEDULE,
    HIGH_WATERMARK_ALLOWED_FIELD_TYPES,
    NO_PARAMETER_OPERATORS,
    RANGE_OPERATORS,
    SINGLE_VALUE_OPERATORS,
    AssertionIncidentBehaviorInputTypes,
    AssertionInfoInputType,
    DetectionMechanismInputTypes,
    FieldSpecType,
    _AllRowsQuery,
    _AllRowsQueryDataHubDatasetProfile,
    _AssertionInput,
    _ChangedRowsQuery,
    _DatasetProfile,
    _try_parse_and_validate_schema_classes_enum,
    get_gms_type_if_criteria_unchanged,
    get_gms_types_if_criteria_unchanged,
)
from acryl_datahub_cloud.sdk.assertion_input.column_assertion_utils import (
    _is_no_parameter_operator,
    _is_range_required_for_operator,
    _is_value_required_for_operator,
    _try_parse_and_validate_range,
    _try_parse_and_validate_range_type,
    _try_parse_and_validate_value,
    _try_parse_and_validate_value_type,
)
from acryl_datahub_cloud.sdk.assertion_input.column_metric_constants import (
    ALLOWED_COLUMN_TYPES_FOR_COLUMN_METRIC_ASSERTION,
    FIELD_METRIC_OPERATOR_CONFIG,
    FIELD_METRIC_TYPE_CONFIG,
    MetricInputType,
    OperatorInputType,
    RangeInputType,
    RangeTypeInputType,
    ValueInputType,
    ValueType,
    ValueTypeInputType,
)
from acryl_datahub_cloud.sdk.entities.assertion import TagsInputType
from acryl_datahub_cloud.sdk.errors import (
    SDKNotYetSupportedError,
    SDKUsageError,
)
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import AssertionUrn, CorpUserUrn, DatasetUrn
from datahub.sdk.entity_client import EntityClient

# New unified criteria parameters type
ColumnMetricAssertionParameters = Union[
    None,  # For operators that don't require parameters (NULL, NOT_NULL)
    ValueInputType,  # Single value
    RangeInputType,  # Range as tuple
]

DEFAULT_DETECTION_MECHANISM_COLUMN_METRIC_ASSERTION: _AllRowsQuery = _AllRowsQuery()


class _ColumnMetricAssertionInput(_AssertionInput):
    """
    Input used to create a column metric assertion.

    This assertion is used to validate the value of a common field / column metric (e.g. aggregation) such as null count + percentage,
    min, max, median, and more.

    Example using the entity models, not comprehensive for all options:

    ```python
    models.AssertionInfoClass(
        type=models.AssertionTypeClass.FIELD,
        fieldAssertion=FieldAssertionInfoClass(
            type=models.FieldAssertionTypeClass.FIELD_METRIC,
            entity=str(self.dataset_urn),
            filter=DatasetFilterClass(
                type=models.DatasetFilterTypeClass.SQL,
                sql="SELECT * FROM dataset WHERE column_name = 'value'",  # Example filter
            ),
            fieldMetricAssertion=FieldMetricAssertionClass(
                field=SchemaFieldSpecClass(
                    path="column_name",  # The column name to validate
                    type="string",  # The type of the column
                    nativeType="string",  # The native type of the column
                ),
                metric=models.FieldMetricTypeClass.NULL_COUNT_PERCENTAGE,  # The metric to validate
                operator=models.AssertionStdOperatorClass.GREATER_THAN,  # The operator to use
                parameters=models.AssertionStdParametersClass(
                    value=models.AssertionStdParameterClass(
                        value=10,  # The value to validate
                        type=models.AssertionStdParameterTypeClass.NUMBER,  # The type of the value
                    ),
                ),
            ),
        ),
        source=models.AssertionSourceClass(
            type=models.AssertionSourceTypeClass.NATIVE,  # Column metric assertions are of type native
            created=AuditStampClass(
                time=1717929600,
                actor="urn:li:corpuser:jdoe",  # The actor who created the assertion
            ),
        ),
        lastUpdated=AuditStampClass(
            time=1717929600,
            actor="urn:li:corpuser:jdoe",  # The actor who last updated the assertion
        ),
        description="This assertion validates the null count percentage of the column 'column_name' is greater than 10.",  # Optional description of the assertion
    )
    ```

    ```python
    models.MonitorInfoClass(
        type=models.MonitorTypeClass.ASSERTION,
        status=models.MonitorStatusClass(
            mode=models.MonitorModeClass.ACTIVE,  # Active or Inactive
        ),
        assertionMonitor=AssertionMonitorClass(
            assertions=AssertionEvaluationSpecClass(
                assertion="urn:li:assertion:123",  # The assertion to monitor
                schedule=models.CronScheduleClass(
                    cron="0 0 * * *",  # The cron schedule
                    timezone="America/New_York",  # The timezone
                ),
                parameters=models.AssertionEvaluationParametersClass(
                    type=models.AssertionEvaluationParametersTypeClass.DATASET_FIELD,
                    datasetFieldParameters=models.DatasetFieldAssertionParametersClass(
                        sourceType=models.DatasetFieldAssertionSourceTypeClass.CHANGED_ROWS_QUERY,  # This can be ALL_ROWS_QUERY, CHANGED_ROWS_QUERY or DATAHUB_DATASET_PROFILE
                        changedRowsField=models.FreshnessFieldSpecClass(
                            path="column_name",
                            type="string",
                            nativeType="string",
                            kind=models.FreshnessFieldKindClass.HIGH_WATERMARK,  # This can be LAST_MODIFIED or HIGH_WATERMARK
                        ),
                    ),
                ),
            ),
        ),
    )
    ```
    """

    def __init__(
        self,
        *,
        # Required parameters
        dataset_urn: Union[str, DatasetUrn],
        entity_client: EntityClient,
        column_name: str,
        metric_type: MetricInputType,
        operator: OperatorInputType,
        # Criteria parameters
        criteria_parameters: Optional[ColumnMetricAssertionParameters] = None,
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: bool = True,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        created_by: Union[str, CorpUserUrn],
        created_at: datetime,
        updated_by: Union[str, CorpUserUrn],
        updated_at: datetime,
        gms_criteria_type_info: Optional[tuple] = None,
        time_bucketing_strategy=None,
    ):
        """
        Initialize a column metric assertion input.

        Args:
            dataset_urn: The dataset urn.
            entity_client: The entity client.
            column_name: The name of the column to validate.
            metric_type: The metric type to validate.
            operator: The operator to use.
            criteria_parameters: The criteria parameters (single value, range tuple, or None). Type will be automatically inferred.
            urn: The urn of the assertion.
            display_name: The display name of the assertion.
            enabled: Whether the assertion is enabled.
            schedule: The schedule of the assertion.
            detection_mechanism: The detection mechanism of the assertion.
            incident_behavior: The incident behavior of the assertion. Accepts strings, enum values, lists, or None.
            tags: The tags of the assertion.
            created_by: The creator of the assertion.
            created_at: The creation time of the assertion.
            updated_by: The updater of the assertion.
            updated_at: The update time of the assertion.
        """
        # Parent will handle validation of common parameters:
        _AssertionInput.__init__(
            self,
            dataset_urn=dataset_urn,
            entity_client=entity_client,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            schedule=schedule,
            detection_mechanism=detection_mechanism,
            incident_behavior=incident_behavior,
            tags=tags,
            source_type=models.AssertionSourceTypeClass.NATIVE,
            created_by=created_by,
            created_at=created_at,
            updated_by=updated_by,
            updated_at=updated_at,
            default_detection_mechanism=DEFAULT_DETECTION_MECHANISM_COLUMN_METRIC_ASSERTION,
            time_bucketing_strategy=time_bucketing_strategy,
        )

        # Column metric assertions (non-smart) don't use exclusion_windows, sensitivity or training_data_lookback_days

        # Validate Column Metric Assertion specific parameters
        self.metric_type = _try_parse_and_validate_schema_classes_enum(
            metric_type, models.FieldMetricTypeClass
        )
        self.column_name = self._try_parse_and_validate_column_name_is_valid_type(
            column_name
        )
        self.operator = _try_parse_and_validate_schema_classes_enum(
            operator, models.AssertionStdOperatorClass
        )

        # Initialize instance variables with proper type annotations
        self.criteria_parameters: Optional[ColumnMetricAssertionParameters] = None
        self.criteria_type: Optional[Union[ValueTypeInputType, RangeTypeInputType]] = (
            None
        )

        # Process criteria parameters with GMS type information if available
        if gms_criteria_type_info is not None:
            self._process_criteria_parameters_with_gms_type(
                criteria_parameters, gms_criteria_type_info
            )
        else:
            self._process_criteria_parameters(criteria_parameters)

        # Validate compatibility:
        self._validate_field_type_and_operator_compatibility(
            self.column_name, self.operator
        )
        self._validate_field_type_and_metric_type_compatibility(
            self.column_name, self.metric_type
        )

    def _infer_criteria_type_from_parameters(
        self,
        criteria_parameters: Optional[ColumnMetricAssertionParameters],
    ) -> Optional[Union[ValueTypeInputType, RangeTypeInputType]]:
        """
        Infer the criteria type from the parameters based on Python types.

        Args:
            criteria_parameters: The criteria parameters to infer type from.

        Returns:
            The inferred type(s) for the criteria parameters.
        """
        if criteria_parameters is None:
            return None

        if isinstance(criteria_parameters, tuple):
            # Range parameters - infer type for each value
            if len(criteria_parameters) != 2:
                raise SDKUsageError(
                    "Range parameters must be a tuple of exactly 2 values"
                )

            type1 = self._infer_single_value_type(criteria_parameters[0])
            type2 = self._infer_single_value_type(criteria_parameters[1])
            return (type1, type2)
        else:
            # Single value parameter
            return self._infer_single_value_type(criteria_parameters)

    def _infer_single_value_type(self, value: ValueInputType) -> ValueTypeInputType:
        """
        Infer the type of a single value based on its Python type.

        Args:
            value: The value to infer type from.

        Returns:
            The inferred ValueType.
        """
        if isinstance(value, (int, float)):
            return ValueType.NUMBER
        elif isinstance(value, str):
            return ValueType.STRING
        else:
            # Default fallback
            return ValueType.UNKNOWN

    def _process_criteria_parameters_with_gms_type(
        self,
        criteria_parameters: Optional[ColumnMetricAssertionParameters],
        gms_type_info: Optional[tuple],
    ) -> None:
        """Process criteria_parameters, using GMS type only if criteria is unchanged."""
        if criteria_parameters is None:
            self._process_none_parameters()
        elif isinstance(criteria_parameters, tuple):
            # Only use GMS types if the user hasn't changed the criteria values
            explicit_types = get_gms_types_if_criteria_unchanged(
                criteria_parameters, gms_type_info
            )
            if explicit_types:
                self._process_range_parameters_with_types(
                    criteria_parameters, explicit_types
                )
            else:
                self._process_range_parameters(criteria_parameters)
        else:
            # Only use GMS type if the user hasn't changed the criteria value
            explicit_type = get_gms_type_if_criteria_unchanged(
                criteria_parameters, gms_type_info
            )
            if explicit_type:
                self._process_single_value_parameters_with_type(
                    criteria_parameters, explicit_type
                )
            else:
                self._process_single_value_parameters(criteria_parameters)

    def _process_criteria_parameters(
        self,
        criteria_parameters: Optional[ColumnMetricAssertionParameters],
    ) -> None:
        """Process the new consolidated criteria_parameters with automatic type inference."""
        if criteria_parameters is None:
            self._process_none_parameters()
        elif isinstance(criteria_parameters, tuple):
            self._process_range_parameters(criteria_parameters)
        else:
            self._process_single_value_parameters(criteria_parameters)

    def _process_none_parameters(self) -> None:
        """Process None criteria_parameters."""
        # No parameters - validation is now handled at the client level
        # This allows both creation and update scenarios to be handled appropriately
        self.criteria_parameters = None
        self.criteria_type = None

    def _process_range_parameters(self, criteria_parameters: tuple) -> None:
        """Process tuple criteria_parameters for range operators."""
        # Range parameters
        if not _is_range_required_for_operator(self.operator):
            raise SDKUsageError(
                f"Operator {self.operator} does not support range parameters. "
                "Provide a single value instead of a tuple."
            )

        # Infer range type automatically
        inferred_range_type = self._infer_criteria_type_from_parameters(
            criteria_parameters
        )

        # Validate and parse the range type
        validated_range_type = _try_parse_and_validate_range_type(inferred_range_type)

        # Validate and parse the range values
        validated_range = _try_parse_and_validate_range(
            criteria_parameters, validated_range_type, self.operator
        )

        # Store validated parameters
        self.criteria_parameters = validated_range
        self.criteria_type = validated_range_type

    def _process_single_value_parameters(
        self, criteria_parameters: Union[str, int, float]
    ) -> None:
        """Process single value criteria_parameters."""
        # Single value parameters
        if _is_no_parameter_operator(self.operator):
            raise SDKUsageError(
                f"Value parameters should not be provided for operator {self.operator}"
            )
        if not _is_value_required_for_operator(self.operator):
            raise SDKUsageError(
                f"Operator {self.operator} does not support value parameters. "
                "Use criteria_parameters=None or omit criteria_parameters."
            )

        # Infer value type automatically
        inferred_value_type = self._infer_criteria_type_from_parameters(
            criteria_parameters
        )

        # Validate value if required
        if _is_value_required_for_operator(self.operator):
            # Validate and parse the value type - make sure it's a single type, not a tuple
            if isinstance(inferred_value_type, tuple):
                raise SDKUsageError("Single value type expected, not a tuple type")

            validated_value_type = _try_parse_and_validate_value_type(
                inferred_value_type
            )
            validated_value = _try_parse_and_validate_value(
                criteria_parameters, validated_value_type
            )

            # Store validated parameters
            self.criteria_parameters = validated_value
            self.criteria_type = validated_value_type
        else:
            # Store raw parameters for operators that don't require validation
            self.criteria_parameters = criteria_parameters
            self.criteria_type = inferred_value_type

    def _process_single_value_parameters_with_type(
        self,
        criteria_parameters: Union[str, int, float],
        gms_type: models.AssertionStdParameterTypeClass,
    ) -> None:
        """Process single value criteria_parameters using explicit GMS type information."""
        # Single value parameters
        if _is_no_parameter_operator(self.operator):
            raise SDKUsageError(
                f"Value parameters should not be provided for operator {self.operator}"
            )
        if not _is_value_required_for_operator(self.operator):
            raise SDKUsageError(
                f"Operator {self.operator} does not support value parameters. "
                "Use criteria_parameters=None or omit criteria_parameters."
            )

        # Use GMS type instead of inferring
        validated_value_type = _try_parse_and_validate_value_type(gms_type)
        validated_value = _try_parse_and_validate_value(
            criteria_parameters, validated_value_type
        )

        # Store validated parameters
        self.criteria_parameters = validated_value
        self.criteria_type = validated_value_type

    def _process_range_parameters_with_types(
        self,
        criteria_parameters: tuple,
        gms_types: tuple,
    ) -> None:
        """Process range criteria_parameters using explicit GMS type information."""
        # Range parameters with GMS types
        if _is_no_parameter_operator(self.operator):
            raise SDKUsageError(
                f"Range parameters should not be provided for operator {self.operator}"
            )
        if not _is_range_required_for_operator(self.operator):
            raise SDKUsageError(
                f"Operator {self.operator} does not support range parameters. "
                "Use a single value or criteria_parameters=None."
            )

        if len(criteria_parameters) != 2:
            raise SDKUsageError("Range parameters must be a tuple of exactly 2 values")

        min_value, max_value = criteria_parameters
        min_type, max_type = gms_types

        # Use GMS types instead of inferring
        validated_min_type = _try_parse_and_validate_value_type(min_type)
        validated_max_type = _try_parse_and_validate_value_type(max_type)

        validated_min_value = _try_parse_and_validate_value(
            min_value, validated_min_type
        )
        validated_max_value = _try_parse_and_validate_value(
            max_value, validated_max_type
        )

        # Store validated parameters
        self.criteria_parameters = (validated_min_value, validated_max_value)
        self.criteria_type = (validated_min_type, validated_max_type)

    def _create_monitor_info(
        self,
        assertion_urn: AssertionUrn,
        status: models.MonitorStatusClass,
        schedule: models.CronScheduleClass,
    ) -> models.MonitorInfoClass:
        """
        Create a MonitorInfoClass with all the necessary components.
        """
        source_type, field = self._convert_assertion_source_type_and_field()
        return models.MonitorInfoClass(
            type=models.MonitorTypeClass.ASSERTION,
            status=status,
            assertionMonitor=models.AssertionMonitorClass(
                assertions=[
                    models.AssertionEvaluationSpecClass(
                        assertion=str(assertion_urn),
                        schedule=schedule,
                        parameters=self._get_assertion_evaluation_parameters(
                            str(source_type), field
                        ),
                    ),
                ],
                settings=None,
                bootstrapConfig=None,
            ),
        )

    def _create_assertion_info(
        self, filter: Optional[models.DatasetFilterClass]
    ) -> AssertionInfoInputType:
        """
        Create a FieldAssertionInfoClass for a column metric assertion.

        Args:
            filter: Optional filter to apply to the assertion.

        Returns:
            A FieldAssertionInfoClass configured for column metric.
        """
        # Get the field spec for the column
        field_spec = self._get_schema_field_spec(self.column_name)

        # Create the field metric assertion
        field_metric_assertion = models.FieldMetricAssertionClass(
            field=field_spec,
            metric=self.metric_type,
            operator=self.operator,
            parameters=self._create_assertion_parameters(),
        )

        # Create the field assertion info
        return models.FieldAssertionInfoClass(
            type=models.FieldAssertionTypeClass.FIELD_METRIC,
            entity=str(self.dataset_urn),
            filter=filter,
            fieldMetricAssertion=field_metric_assertion,
            fieldValuesAssertion=None,  # Explicitly set to None since this is a field metric assertion
        )

    def _convert_schedule(self) -> models.CronScheduleClass:
        """
        Create a schedule for a column metric assertion.

        Returns:
            A CronScheduleClass with appropriate schedule settings.
        """
        if self.schedule is None:
            return DEFAULT_DAILY_SCHEDULE

        return models.CronScheduleClass(
            cron=self.schedule.cron,
            timezone=self.schedule.timezone,
        )

    def _convert_schema_field_spec_to_freshness_field_spec(
        self, field_spec: models.SchemaFieldSpecClass
    ) -> models.FreshnessFieldSpecClass:
        """
        Convert a SchemaFieldSpecClass to a FreshnessFieldSpecClass.
        """
        return models.FreshnessFieldSpecClass(
            path=field_spec.path,
            type=field_spec.type,
            nativeType=field_spec.nativeType,
            kind=models.FreshnessFieldKindClass.HIGH_WATERMARK,
        )

    def _get_assertion_evaluation_parameters(
        self, source_type: str, field: Optional[FieldSpecType]
    ) -> models.AssertionEvaluationParametersClass:
        """
        Get evaluation parameters for a column metric assertion.
        Converts SchemaFieldSpecClass to FreshnessFieldSpecClass if needed.
        """
        if field is not None:
            if isinstance(field, models.SchemaFieldSpecClass):
                field = self._convert_schema_field_spec_to_freshness_field_spec(field)
            assert isinstance(field, models.FreshnessFieldSpecClass), (
                "Field must be FreshnessFieldSpecClass for monitor info"
            )
        return models.AssertionEvaluationParametersClass(
            type=models.AssertionEvaluationParametersTypeClass.DATASET_FIELD,
            datasetFieldParameters=models.DatasetFieldAssertionParametersClass(
                sourceType=source_type,
                changedRowsField=field,
                timeBucketingStrategy=self.time_bucketing_strategy,
            ),
        )

    def _convert_assertion_source_type_and_field(
        self,
    ) -> tuple[str, Optional[FieldSpecType]]:
        """
        Convert detection mechanism into source type and field specification for column metric assertions.

        Returns:
            A tuple of (source_type, field) where field may be None.
            Note that the source_type is a string, not a models.DatasetFieldAssertionSourceTypeClass (or other assertion source type) since
            the source type is not a enum in the code generated from the DatasetFieldSourceType enum in the PDL.

        Raises:
            SDKNotYetSupportedError: If the detection mechanism is not supported.
            SDKUsageError: If the field (column) is not found in the dataset,
            and the detection mechanism requires a field. Also if the field
            is not an allowed type for the detection mechanism.
        """
        source_type = models.DatasetFieldAssertionSourceTypeClass.ALL_ROWS_QUERY
        field = None
        SUPPORTED_DETECTION_MECHANISMS = [
            _AllRowsQuery().type,
            _AllRowsQueryDataHubDatasetProfile().type,
            _ChangedRowsQuery(column_name="").type,
        ]

        if isinstance(self.detection_mechanism, _ChangedRowsQuery):
            source_type = models.DatasetFieldAssertionSourceTypeClass.CHANGED_ROWS_QUERY
            column_name = self._try_parse_and_validate_column_name_is_valid_type(
                self.detection_mechanism.column_name,  # The high watermark column name
                allowed_column_types=HIGH_WATERMARK_ALLOWED_FIELD_TYPES,
            )
            field = self._get_schema_field_spec(column_name)
        elif isinstance(self.detection_mechanism, _AllRowsQuery):
            source_type = models.DatasetFieldAssertionSourceTypeClass.ALL_ROWS_QUERY
            # For query-based detection, we don't need a field specification
            # as the query itself defines what data to analyze
        elif isinstance(
            self.detection_mechanism,
            (_AllRowsQueryDataHubDatasetProfile, _DatasetProfile),
        ):
            source_type = (
                models.DatasetFieldAssertionSourceTypeClass.DATAHUB_DATASET_PROFILE
            )
            # Note: This is only valid on the all rows query
        else:
            raise SDKNotYetSupportedError(
                f"Detection mechanism {self.detection_mechanism} is not supported for column metric assertions, please use a supported detection mechanism: {', '.join(SUPPORTED_DETECTION_MECHANISMS)}"
            )

        return source_type, field

    def _create_assertion_parameters(self) -> models.AssertionStdParametersClass:
        """
        Create assertion parameters based on the operator type and provided values.

        Returns:
            An AssertionStdParametersClass with the appropriate parameters.

        Raises:
            SDKUsageError: If the parameters are invalid for the operator type.
        """
        if self.operator in SINGLE_VALUE_OPERATORS:
            if self.criteria_parameters is None or isinstance(
                self.criteria_parameters, tuple
            ):
                raise SDKUsageError(
                    f"Single value is required for operator {self.operator}"
                )
            if self.criteria_type is None or isinstance(self.criteria_type, tuple):
                raise SDKUsageError(
                    f"Single value type is required for operator {self.operator}"
                )
            return models.AssertionStdParametersClass(
                value=models.AssertionStdParameterClass(
                    value=str(self.criteria_parameters),
                    type=self.criteria_type,
                ),
            )
        elif self.operator in RANGE_OPERATORS:
            if not isinstance(self.criteria_parameters, tuple):
                raise SDKUsageError(
                    f"Range parameters are required for operator {self.operator}"
                )
            if not isinstance(self.criteria_type, tuple):
                raise SDKUsageError(
                    f"Range type is required for operator {self.operator}"
                )
            return models.AssertionStdParametersClass(
                minValue=models.AssertionStdParameterClass(
                    value=str(self.criteria_parameters[0]),
                    type=self.criteria_type[0],
                ),
                maxValue=models.AssertionStdParameterClass(
                    value=str(self.criteria_parameters[1]),
                    type=self.criteria_type[1],
                ),
            )
        elif self.operator in NO_PARAMETER_OPERATORS:
            return models.AssertionStdParametersClass()
        else:
            raise SDKUsageError(f"Unsupported operator type: {self.operator}")

    def _try_parse_and_validate_column_name_is_valid_type(
        self,
        column_name: str,
        allowed_column_types: list[
            models.DictWrapper
        ] = ALLOWED_COLUMN_TYPES_FOR_COLUMN_METRIC_ASSERTION,
    ) -> str:
        """
        Parse and validate a column name. Determine from the field spec if the column exists and is of the appropriate type for the metric type.
        Validate that this is a column that is valid for the metric type, see also getEligibleFieldColumns and related functions in the frontend
        """
        field_spec = self._get_schema_field_spec(column_name)
        self._validate_field_type(
            field_spec,
            column_name,
            allowed_column_types,
            "column metric assertion",
        )
        return column_name

    def _assertion_type(self) -> str:
        """Get the assertion type."""
        return models.AssertionTypeClass.FIELD

    def _validate_field_type_and_operator_compatibility(
        self, column_name: str, operator: models.AssertionStdOperatorClass
    ) -> None:
        """Validate that the operator is compatible with column metric assertions.

        Column metric assertions always produce numeric results, regardless of the underlying
        column's data type. For example, NULL_COUNT on a DATE column produces a numeric count,
        not a date. Therefore, operators are validated against FIELD_METRIC_OPERATOR_CONFIG
        (the numeric metric operators) rather than against the column's raw data type.

        Args:
            column_name: The name of the column to validate.
            operator: The operator to validate against.

        Raises:
            SDKUsageError: If the operator is not valid for column metric assertions.
        """
        if operator not in FIELD_METRIC_OPERATOR_CONFIG:
            raise SDKUsageError(
                f"Operator {operator} is not allowed for column metric assertions on column '{column_name}'. "
                f"Allowed operators: {', '.join(str(op) for op in FIELD_METRIC_OPERATOR_CONFIG)}"
            )

    def _validate_field_type_and_metric_type_compatibility(
        self, column_name: str, metric_type: models.FieldMetricTypeClass
    ) -> None:
        """Validate that the metric type is compatible with the field type.

        See FIELD_METRIC_TYPE_CONFIG in the frontend for the allowed metric types for each field type.

        Args:
            column_name: The name of the column to validate.
            metric_type: The metric type to validate.

        Raises:
            SDKUsageError: If the metric type is not compatible with the field type.
        """
        field_spec = self._get_schema_field_spec(column_name)
        field_type = field_spec.type

        if field_type not in FIELD_METRIC_TYPE_CONFIG:
            raise SDKUsageError(
                f"Column {column_name} is of type {field_type}, which is not supported for column metric assertions"
            )

        allowed_metric_types = FIELD_METRIC_TYPE_CONFIG[field_type]
        if metric_type not in allowed_metric_types:
            raise SDKUsageError(
                f"Metric type {metric_type} is not allowed for field type {field_type}. Allowed metric types: {', '.join(str(mt) for mt in allowed_metric_types)}"
            )



import logging
import time
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Tuple

from pydantic import BaseModel
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from acryl_datahub_cloud.datahub_forms_notifications.query import (
    GRAPHQL_GET_FEATURE_FLAG,
    GRAPHQL_GET_SEARCH_RESULTS_TOTAL,
    GRAPHQL_SCROLL_FORMS_FOR_NOTIFICATIONS,
    GRAPHQL_SEND_FORM_NOTIFICATION_REQUEST,
)
from acryl_datahub_cloud.graphql_utils import parse_extra_properties_for_model
from acryl_datahub_cloud.notifications.notification_recipient_builder import (
    NotificationRecipientBuilder,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.graph.filters import RawSearchFilter
from datahub.metadata.schema_classes import (
    FormInfoClass,
    FormNotificationDetailsClass,
    FormNotificationEntryClass,
    FormNotificationsClass,
    FormSettingsClass,
    FormStateClass,
    FormTypeClass,
)

logger = logging.getLogger(__name__)

USER_URN_PREFIX = "urn:li:corpuser"
GROUP_URN_PREFIX = "urn:li:corpGroup"


class DataHubFormsNotificationsSourceConfig(BaseModel):
    form_urns: Optional[List[str]] = None


class DataHubDatasetSearchRow(BaseModel):
    urn: str
    owners: List[str] = []


@dataclass
class DataHubFormsNotificationsSourceReport(SourceReport):
    notifications_sent: int = (
        0  # the number of recipients we sent notifications out for
    )
    forms_count: int = (
        0  # the number of forms that we sent at least one nitification for
    )


@platform_name(id="datahub", platform_name="DataHub")
@config_class(DataHubFormsNotificationsSourceConfig)
@support_status(SupportStatus.INCUBATING)
class DataHubFormsNotificationsSource(Source):
    """Forms Notification Source that notifies recipients for compliance forms tasks"""

    def __init__(
        self, config: DataHubFormsNotificationsSourceConfig, ctx: PipelineContext
    ):
        super().__init__(ctx)
        self.config: DataHubFormsNotificationsSourceConfig = config
        self.report = DataHubFormsNotificationsSourceReport()
        self.graph: DataHubGraph = ctx.require_graph(
            "Loading default graph coordinates."
        )
        self.group_to_users_map: Dict[str, List[str]] = {}
        self.recipient_builder: NotificationRecipientBuilder = (
            NotificationRecipientBuilder(self.graph)
        )
        self.user_to_form_notifications: Dict[str, FormNotificationsClass] = {}

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        # end early if the feature flag is not enabled
        if not self.is_feature_flag_enabled():
            return []

        self.notify_form_assignees()

        # This source doesn't produce any work units
        return []

    def is_feature_flag_enabled(self) -> bool:
        response = self.execute_graphql_with_retry(
            GRAPHQL_GET_FEATURE_FLAG, variables={}
        )

        result = response.get("appConfig", {})
        featureFlags = result.get("featureFlags", {})
        is_enabled = featureFlags.get("formsNotificationsEnabled", False)

        if not is_enabled:
            logger.error(
                "Tried running datahub-forms-notifications with formsNotificationsEnabled disabled"
            )

        return is_enabled

    def notify_form_assignees(self) -> None:
        for urn, form in self.get_forms():
            if not self.is_form_complete(urn, form.type):
                assignees = self.get_form_assignees(urn, form)
                self.process_notify_on_publish(
                    assignees, form.name, urn, form.description
                )

    def process_notify_on_publish(
        self,
        form_assignees: List[str],
        form_name: str,
        form_urn: str,
        form_details: str | None,
    ) -> None:
        """
        Take in form assignees, find the ones who haven't been notified on publish, and build a notification for them.
        """
        filtered_assignees = self.filter_assignees_to_notify(form_assignees, form_urn)
        recipients = []
        if self.recipient_builder is not None:
            recipients = self.recipient_builder.build_actor_recipients(
                filtered_assignees, "COMPLIANCE_FORM_PUBLISH", True
            )
        recipient_count = len(recipients)

        if recipient_count > 0:
            self.report.notifications_sent += recipient_count
            self.report.forms_count += 1

            parameters = [{"key": "formName", "value": form_name}]
            if form_details is not None:
                parameters.append({"key": "formDetails", "value": form_details})

            response = self.execute_graphql_with_retry(
                GRAPHQL_SEND_FORM_NOTIFICATION_REQUEST,
                variables={
                    "input": {
                        "type": "BROADCAST_COMPLIANCE_FORM_PUBLISH",
                        "parameters": parameters,
                        "recipients": self.recipient_builder.convert_recipients_to_json_objects(
                            recipients
                        ),
                    }
                },
            )

            if not response.get("sendFormNotificationRequest", False):
                logger.error(
                    f"Issue sending the notification request for this job. Response: {response}"
                )
            else:
                unique_actor_urns = set(
                    [
                        recipient.get("actor")
                        for recipient in recipients
                        if recipient.get("actor") is not None
                    ]
                )
                for actor_urn in unique_actor_urns:
                    self.update_form_notifications(actor_urn, form_urn)

    def update_form_notifications(self, user_urn: str, form_urn: str) -> None:
        """
        After sending a notification, update the user's formNotifications aspect
        to track that we sent the notification that we did
        """
        # get or create default formNotifications aspect
        form_notifications = self.user_to_form_notifications.get(user_urn)
        if form_notifications is None:
            form_notifications = FormNotificationsClass(notificationDetails=[])

        # get the notification details for our specific form or create default
        details_for_form = self.get_notification_details_for_form(
            user_urn, form_urn, form_notifications
        )
        if details_for_form is None:
            details_for_form = FormNotificationDetailsClass(
                formUrn=form_urn, notificationLog=[]
            )

        # add new notification log entry for this occasion
        new_notification_log_entry = FormNotificationEntryClass(
            time=int(time.time() * 1000),
            notificationType="BROADCAST_COMPLIANCE_FORM_PUBLISH",
        )
        details_for_form.notificationLog.append(new_notification_log_entry)

        # filter out details for given form so we can add updated one
        final_notification_details = [
            details
            for details in form_notifications.notificationDetails
            if details.formUrn != form_urn
        ]
        final_notification_details.append(details_for_form)

        # update the aspect with the final notification details
        form_notifications.notificationDetails = final_notification_details

        self.user_to_form_notifications[user_urn] = form_notifications

        self.graph.emit(
            MetadataChangeProposalWrapper(
                entityUrn=user_urn,
                aspect=form_notifications,
            )
        )

    @retry(
        retry=retry_if_exception_type((Exception, ConnectionError)),
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        reraise=True,
    )
    def execute_graphql_with_retry(
        self, query: str, variables: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Execute GraphQL query with retry logic"""
        if self.graph is None:
            raise ValueError("Graph client not initialized")
        response = self.graph.execute_graphql(query, variables=variables)
        error = response.get("error")
        if error:
            raise Exception(f"GraphQL error: {error}")
        return response

    def get_forms(self) -> List[Tuple[str, FormInfoClass]]:
        """
        Get forms and their formInfo aspect either from the forms provided in the config
        or search for forms that are published and notifyAssigneesOnPublish = True.
        This method will only return forms that are published and have notifications enabled.
        """
        form_urns = []

        if self.config.form_urns is not None:
            form_urns = self.config.form_urns
        else:
            form_urns = self.search_for_forms()

        form_urns_with_notifications_enabled = (
            self.get_form_urns_with_notifications_enabled(form_urns)
        )

        return self.get_form_infos(form_urns_with_notifications_enabled)

    def get_form_urns_with_notifications_enabled(
        self, form_urns: List[str]
    ) -> List[str]:
        """
        Get formSettings aspects and check if notifications are enabled for a given form urn.
        If notifications are enabled, add to filtered list and return.
        """
        filtered_form_urns: List[str] = []

        if len(form_urns) > 0:
            entities = self.graph.get_entities("form", form_urns, ["formSettings"])
            for urn, entity in entities.items():
                form_tuple = entity.get(FormSettingsClass.ASPECT_NAME, (None, None))
                if form_tuple and form_tuple[0]:
                    if not isinstance(form_tuple[0], FormSettingsClass):
                        logger.error(
                            f"{form_tuple[0]} is not of type FormInfo for urn: {urn}"
                        )
                    else:
                        form_settings = form_tuple[0]
                        if form_settings.notificationSettings.notifyAssigneesOnPublish:
                            filtered_form_urns.append(urn)

        return filtered_form_urns

    def get_form_infos(self, form_urns: List[str]) -> List[Tuple[str, FormInfoClass]]:
        """
        Get formInfo aspects for a list of form urns and return the formInfos of forms
        that are published. If a form is not published, we don't want to notify.
        """
        form_infos: List[Tuple[str, FormInfoClass]] = []

        if len(form_urns) > 0:
            entities = self.graph.get_entities("form", form_urns, ["formInfo"])
            for urn, entity in entities.items():
                form_tuple = entity.get(FormInfoClass.ASPECT_NAME, (None, None))
                if form_tuple and form_tuple[0]:
                    if not isinstance(form_tuple[0], FormInfoClass):
                        logger.error(
                            f"{form_tuple[0]} is not of type FormInfo for urn: {urn}"
                        )
                    else:
                        form_info = form_tuple[0]
                        if form_info.status.state == FormStateClass.PUBLISHED:
                            form_infos.append((urn, form_tuple[0]))

        return form_infos

    def search_for_forms(self) -> List[str]:
        scroll_id: Optional[str] = None
        form_urns: List[str] = []

        try:
            while True:
                next_scroll_id, results = self.scroll_forms_to_notify_for(scroll_id)

                for result in results:
                    form_urn = result.get("entity", {}).get("urn", None)
                    if form_urn is None:
                        self.report.report_warning(
                            message="Failed to resolve entity urn for form! Skipping...",
                            context=f"Response: {str(result)}",
                        )
                    else:
                        form_urns.append(form_urn)

                if next_scroll_id is None:
                    break
                else:
                    scroll_id = next_scroll_id

                time.sleep(1)

        except Exception as e:
            self.report.report_failure(
                title="Failed to search for forms to send notifications for",
                message="Error occurred while searching for forms to send notifications for",
                context=f"message = {str(e)}",
                exc=e,
            )
            return form_urns

        return form_urns

    def scroll_forms_to_notify_for(
        self, scroll_id: Optional[str]
    ) -> Tuple[Optional[str], List[Dict[str, Any]]]:
        """Scroll through shared entities with retry logic"""
        response = self.execute_graphql_with_retry(
            GRAPHQL_SCROLL_FORMS_FOR_NOTIFICATIONS,
            variables={
                "scrollId": scroll_id,
                "count": 500,
            },
        )

        result = response.get("scrollAcrossEntities", {})
        return result.get("nextScrollId"), result.get("searchResults", [])

    def get_form_assignees(self, form_urn: str, form: FormInfoClass) -> List[str]:
        """
        Form assignees are provided explicitly on the form and the owners of assets with this form
        if it's an ownership form.
        For form notifications, we want to get users from a user group and send notifications to
        those users specifically
        """
        user_urns = form.actors.users if form.actors.users is not None else []
        group_urns = form.actors.groups if form.actors.groups is not None else []

        if form.actors.owners:
            (user_owners, group_owners) = self.get_owners_of_assets_for_form(
                form_urn, form
            )
            user_urns.extend(user_owners)
            group_urns.extend(group_owners)

        for group_urn in group_urns:
            user_urns.extend(self._get_users_in_group(group_urn))

        return list(set(user_urns))

    def get_owners_of_assets_for_form(
        self, form_urn: str, form: FormInfoClass
    ) -> Tuple[List[str], List[str]]:
        """
        Filter to get assets that are not complete for this form and using the extra_source_fields parameter
        we pull owners from the asset's elastic row. self.graph.get_results_by_filter will paginate over assets
        """
        user_urns = []
        group_urns = []

        extra_fields = [f for f in DataHubDatasetSearchRow.model_fields]
        results = self.graph.get_results_by_filter(
            extra_or_filters=self._get_incomplete_assets_for_form(form_urn, form.type),
            extra_source_fields=extra_fields,
            skip_cache=True,
        )
        for result in results:
            extra_properties = result["extraProperties"]
            extra_properties_map = parse_extra_properties_for_model(
                extra_properties, DataHubDatasetSearchRow
            )
            search_row = DataHubDatasetSearchRow(**extra_properties_map)
            for owner in search_row.owners:
                if owner.startswith(USER_URN_PREFIX):
                    user_urns.append(owner)
                elif owner.startswith(GROUP_URN_PREFIX):
                    group_urns.append(owner)
                else:
                    logger.warning(
                        f"Found unexpected owner {owner} for asset {search_row.urn}"
                    )

        return (user_urns, group_urns)

    def filter_assignees_to_notify(
        self, user_urns: List[str], form_urn: str
    ) -> List[str]:
        """
        Filter out any users who have already received the publish notification type in the past
        """
        filtered_users = []

        self.populate_user_to_form_notifications(user_urns)

        for user in user_urns:
            form_notifications = self.user_to_form_notifications.get(user)
            if form_notifications is None or not self.has_user_been_sent_notification(
                user, form_urn, form_notifications, "BROADCAST_COMPLIANCE_FORM_PUBLISH"
            ):
                filtered_users.append(user)

        return filtered_users

    def has_user_been_sent_notification(
        self,
        user_urn: str,
        form_urn: str,
        form_notifications: FormNotificationsClass,
        notification_type: str,
    ) -> bool:
        notification_details = self.get_notification_details_for_form(
            user_urn, form_urn, form_notifications
        )
        if notification_details is None:
            return False

        notification_types_sent_for_form = [
            entry.notificationType for entry in notification_details.notificationLog
        ]

        return notification_type in notification_types_sent_for_form

    def get_notification_details_for_form(
        self, user_urn: str, form_urn: str, form_notifications: FormNotificationsClass
    ) -> FormNotificationDetailsClass | None:
        notification_details_for_form = [
            detail
            for detail in form_notifications.notificationDetails
            if detail.formUrn == form_urn
        ]

        notification_details = None
        if len(notification_details_for_form) > 1:
            logger.warning(
                f"Found more than one notificationDetails for a given form for user {user_urn} in {form_notifications}"
            )
            # grab first one
            notification_details = notification_details_for_form[0]
        elif len(notification_details_for_form) == 1:
            notification_details = notification_details_for_form[0]

        return notification_details

    def populate_user_to_form_notifications(self, user_urns: List[str]) -> None:
        new_users = [
            urn for urn in user_urns if urn not in self.user_to_form_notifications
        ]

        if len(new_users) == 0:
            return

        entities = self.graph.get_entities("corpuser", new_users, ["formNotifications"])
        for urn, entity in entities.items():
            user_tuple = entity.get(FormNotificationsClass.ASPECT_NAME, (None, None))
            if user_tuple and user_tuple[0]:
                if not isinstance(user_tuple[0], FormNotificationsClass):
                    logger.error(
                        f"{user_tuple[0]} is not of type FormNotifications for urn: {urn}"
                    )
                else:
                    self.user_to_form_notifications[urn] = user_tuple[0]

    def _get_users_in_group(self, group_urn: str) -> List[str]:
        """
        Using a relationship query, get users inside of a group. Store these users in memory if we've
        already fetched the users for this group.
        """
        if (users_in_group := self.group_to_users_map.get(group_urn)) is not None:
            return users_in_group

        group_member_urns = []
        members = self.graph.get_related_entities(
            group_urn,
            ["IsMemberOfGroup", "IsMemberOfNativeGroup"],
            self.graph.RelationshipDirection.INCOMING,
        )
        member_urns = [member.urn for member in members]
        for member_urn in member_urns:
            if member_urn.startswith(USER_URN_PREFIX):
                group_member_urns.append(member_urn)
            else:
                logger.warning(
                    f"Unexpected group member {member_urn} found in group {group_urn}"
                )
        self.group_to_users_map[group_urn] = group_member_urns

        return group_member_urns

    def _get_verification_form_filter(self, form_urn: str) -> RawSearchFilter:
        return [
            {"and": [{"field": "incompleteForms", "values": [form_urn]}]},
            {
                "and": [
                    {"field": "completedForms", "values": [form_urn]},
                    {"field": "verifiedForms", "values": [form_urn], "negated": True},
                ]
            },
        ]

    def _get_completion_form_filter(self, form_urn: str) -> RawSearchFilter:
        return [{"and": [{"field": "incompleteForms", "values": [form_urn]}]}]

    def _get_incomplete_assets_for_form(
        self, form_urn: str, form_type: str | FormTypeClass
    ) -> RawSearchFilter:
        return (
            self._get_completion_form_filter(form_urn)
            if form_type == FormTypeClass.COMPLETION
            else self._get_verification_form_filter(form_urn)
        )

    def is_form_complete(self, form_urn: str, form_type: str | FormTypeClass) -> int:
        """
        Returns whether this form is complete - meaning no assets have any work left to do for it.
        This takes into account the type of form to know if it's fully complete.
        """
        response = self.execute_graphql_with_retry(
            GRAPHQL_GET_SEARCH_RESULTS_TOTAL,
            variables={
                "count": 0,
                "orFilters": self._get_incomplete_assets_for_form(form_urn, form_type),
            },
        )

        result = response.get("searchAcrossEntities", {})
        total = result.get("total", -1)
        if total < 0:
            logger.warning(
                f"Error evaluating if form with urn {form_urn} is complete. Skipping."
            )
            return True
        else:
            return total == 0

    def get_report(self) -> SourceReport:
        return self.report



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import CorpGroupEditableInfoClass
from .....schema_classes import CorpGroupInfoClass
from .....schema_classes import CorpGroupSettingsClass
from .....schema_classes import CorpUserAppearanceSettingsClass
from .....schema_classes import CorpUserCredentialsClass
from .....schema_classes import CorpUserEditableInfoClass
from .....schema_classes import CorpUserHomePageSettingsClass
from .....schema_classes import CorpUserInfoClass
from .....schema_classes import CorpUserInvitationStatusClass
from .....schema_classes import CorpUserSettingsClass
from .....schema_classes import CorpUserStatusClass
from .....schema_classes import CorpUserViewsSettingsClass
from .....schema_classes import GroupMembershipClass
from .....schema_classes import InvitationStatusClass
from .....schema_classes import InviteTokenClass
from .....schema_classes import NativeGroupMembershipClass
from .....schema_classes import RoleMembershipClass
from .....schema_classes import TokenTypeClass
from .....schema_classes import UserAiPluginConfigClass
from .....schema_classes import UserAiPluginSettingsClass
from .....schema_classes import UserApiKeyConnectionConfigClass
from .....schema_classes import UserOAuthConnectionConfigClass


CorpGroupEditableInfo = CorpGroupEditableInfoClass
CorpGroupInfo = CorpGroupInfoClass
CorpGroupSettings = CorpGroupSettingsClass
CorpUserAppearanceSettings = CorpUserAppearanceSettingsClass
CorpUserCredentials = CorpUserCredentialsClass
CorpUserEditableInfo = CorpUserEditableInfoClass
CorpUserHomePageSettings = CorpUserHomePageSettingsClass
CorpUserInfo = CorpUserInfoClass
CorpUserInvitationStatus = CorpUserInvitationStatusClass
CorpUserSettings = CorpUserSettingsClass
CorpUserStatus = CorpUserStatusClass
CorpUserViewsSettings = CorpUserViewsSettingsClass
GroupMembership = GroupMembershipClass
InvitationStatus = InvitationStatusClass
InviteToken = InviteTokenClass
NativeGroupMembership = NativeGroupMembershipClass
RoleMembership = RoleMembershipClass
TokenType = TokenTypeClass
UserAiPluginConfig = UserAiPluginConfigClass
UserAiPluginSettings = UserAiPluginSettingsClass
UserApiKeyConnectionConfig = UserApiKeyConnectionConfigClass
UserOAuthConnectionConfig = UserOAuthConnectionConfigClass

# fmt: on
